/* FFTdisplay.c

	Copyright Larry Cicchinelli (K3PTO) 2006

For the fft analysis the frequency associated with the output can be
determined as follows:
	Freq = k*N where
   	N = the output array index
	   k = Fsample/FFTsize*2 = measurement resolution in Hz/index
   therefore: Freq = N*Fsample/FFTsize*2

Both the LCD and A/D use serial port B in its SPI mode
	PC4 = serial port B Tx
	PC5 = serial port B Rx
	PB0 = serial port B clock
The LCD uses (see F51553.lib):
	PE1 = /ChipSelect
	PE3 = /Reset
	PE5 = A0: 1 = data, 0 = command
   PE7 = PWM for BackLight brightness
The A/D uses:
   PE0 = /ChipSelect
The relays use:
	PD0 = K1
	PD2 = K2
   PD4 = K3
   PD6 = K4
The switches use (open = high).  The function Switches() inverts the bits so
that On = 1.
	PD1 = Sw1
   PD3 = Sw2
   PD5 = Sw3
   PD7 = Sw4
*/
#memmap xmem

#define VERSION	1.0

// define only ONE of the following two values
//#define ANALOG_1
#define ANALOG_2
/******************/

// ************** debugging macros ****************
//#define F51553_DEBUG_asm		debug		// F51553.LIB: enable debug asm functions
//#define F51553_DEBUG			debug		// F51553.LIB: enable debug C functions
//#define F51553_MENU_DEBUG	debug		// F51553menu.LIB: enable menu debugging
#define FFT_DEBUG				debug		// debug this program

#ifndef FFT_DEBUG
	#define FFT_DEBUG nodebug
#endif

#define DEBUG_FFT_DISPLAY 0
			// 1 = initial FFT values
			// 2 = magnitude of FFT values
         // 8 = normalized histogram values
         // 0x10 = raw data
#ifndef DEBUG_FFT_DISPLAY
	#define DEBUG_FFT_DISPLAY 0
#endif

#define DEBUG_PRINT_ADvalues 512
#ifndef DEBUG_PRINT_ADvalues
	#define DEBUG_PRINT_ADvalues 0
#endif

//#define USE_DUMMY_DATA					// fill data array with dummy data

// ************** end debugging macros ****************


// The following macros are based on the I/O as stated above

// LCD defintions *****************************************
//		all LCD control bits MUST in the same register
#define LCD_REG			PEDR
#define LCD_SHADOW		PEDRShadow
#define LCD_DDR			PEDDR
#define LCD_DDR_SHADOW	PEDDRShadow
#define LCD_CS_BIT		1
#define LCD_RESET_BIT 	3
#define LCD_A0_BIT		5
#define LCD_PWM_BIT		7
#define LCD_CS_MASK		(1<<LCD_CS_BIT)
#define LCD_RESET_MASK	(1<<LCD_RESET_BIT)
#define LCD_A0_MASK		(1<<LCD_A0_BIT)
#define LCD_PWM_MASK		(1<<LCD_PWM_BIT)
#define LCD_PWM_CH		(LCD_PWM_BIT-4)

// A/D definitions ****************************************
#define AD_DR				PEDR
#define AD_SHADOW			PEDRShadow
#define AD_DDR				PEDDR
#define AD_DDR_SHADOW	PEDDRShadow
#define AD_CS_BIT			0
#define AD_CS_MASK		(1<<AD_CS_BIT)

// the following two macros can only be used in a #asm block
#define	ADCS_ON						$\
			ld		a, (AD_SHADOW)		$\
   		and	~AD_CS_MASK			$\
   ioi	ld		(AD_DR), a

#define	ADCS_OFF						$\
			ld		a, (AD_SHADOW)		$\
   		or		+AD_CS_MASK			$\
   ioi	ld		(AD_DR), a

// Relay definitions **************************************
#define RELAY_DR		PDDR
#define RELAY_DDR		PDDDR
#define RELAY_SHADOW	PDDRShadow
#define RELAY_DDRSHADOW	PDDDRShadow
#define RLY1			0x01		// bit 0
#define RLY2			0x04		// bit 2
#define RLY3			0x10		// bit 4
#define RLY4			0x40		// bit 6
#define RLY_MASK		(RLY1|RLY2|RLY3|RLY4)
#define MAXGAINS		16


#ifdef ANALOG_1
// Analog board calibration stuff
const char	CalPots[][4] = { "Ref", "R35" };
const char	CalRelays[] = {
   (RLY2|RLY3),	// ref
   RLY2 };			// 20db block

const int GainMask[] = { 0x11, 0x01, 0x14 };
const int dbGainValues[] = { 0, 20, -30 };
#endif

#ifdef ANALOG_2
// Analog board calibration stuff
const char	CalPots[][4] = { "Ref", "R13", "R18", "R23" };
const char	CalRelays[] = {
	(RLY_MASK-RLY1),				// all gain blocks set to 0db, leave atten in
   (RLY_MASK-RLY1-RLY2),		// 1st 20db block
   (RLY_MASK-RLY1-RLY3),		// 2nd 20db block
   (RLY_MASK-RLY1-RLY4) };		// 10db block

// Array of values to enable the different gain blocks - note that the value
// uses inverse logic for the gain blocks: control is 0 to enable gain value
// except for the attenuator.
const char GainMask[] = {
	(RLY_MASK-RLY1),					// -30
	(RLY_MASK-RLY1-RLY4),			// -20
	(RLY_MASK-RLY1-RLY2),			// -10
   (RLY_MASK),							// 0
   (RLY_MASK-RLY4),					// 10
   (RLY_MASK-RLY2),					// 20
   (RLY_MASK-RLY2-RLY4),			// 30
   (RLY_MASK-RLY2-RLY3),			// 40
   (RLY_MASK-RLY2-RLY3-RLY4) };	// 50

// List of gain values - must be same nbr of elements as GainMask
const int dbGainValues[] = { -30, -20, -10, 0, 10, 20, 30, 40, 50 };
#endif

// Switch definitions *************************************
#define SWITCH_REG	PDDR
#define SW1_BIT		1
#define SW2_BIT		3
#define SW3_BIT		5
#define SW4_BIT		7
/*******************************************************************/


#use F51553.lib
#use F51553menu.lib
#use F51553font_2R.Lib
#use F51553smNbrs.Lib

#define	LCD_GRAPH_PAGES	7				// nbr of LCD pages to use for the graph
#define	LCD_GRAPH_HEIGHT	(LCD_GRAPH_PAGES*8)
#define	LCD_GRAPH_LAST_BAR	127	// columns are 0 - 127
#define	LCD_GRAPH_BARS	100
#define	LCD_GRAPH_START_BAR (LCD_GRAPH_LAST_BAR-LCD_GRAPH_BARS+1)

#use fft.lib		// integer FFT.


int	CalcMagnitude ( void );
void	DumpADvalues ( char *title );
void	DummyData ( void );
void	SetInterruptRate ( void );
int	Switches ( void );
int	CkDelay ( long InitTimerValue );
void	TimerISR ( void );
int	db2vr ( int db );						// convert db to voltage ratio
int	vr2db ( int VR );						// convert ratio to db
void	fDispGains ( int Flag );			// display Analog Board gains

void fTestLCD ( void );
void TestIO ( void );

unsigned long int timer;	// used to measure execution time
char	LCDstring[33];
/*******************************************************************/

// these are the values which are stored in the User Block at offset 0
// THE KeyValue MUST BE CHANGED WHEN ANY CHANGE
//		TO THE VALUES STORED IN THE STRUCTURE IS MADE

// the default value for INI_KEY_VALUE should be changed whenever the
// UB structure is changed.
#define	INI_KEY_VALUE     		10002

struct
{
	unsigned KeyValue;
	unsigned	SampleFreq;						// Hz
	int		MaxDispFreq;					// Hz
   int		MinDispFreq;					// Hz
   unsigned	SampleSizeIndex;				// index into SampleSizes
	unsigned	MenuDelay;						// msec
   int		Brightness;						// value
   int		Backlight;						// %
   unsigned Mode;
   unsigned Threshold;
	int		dbRange;
   int		RefHt;
   int		vRange;
   int		ADzero;							// count: no signal A/D value
	int		Cal_mv;
   int		Cal_fft_value[3];				// for sample sizes: 512, 256, 128
   int		GainSel;							// 0 relative
   int		dbGain[MAXGAINS];				// gain values
   int		Relays[MAXGAINS];				// relay pattern for gains
   int		NbrGains;
}UB;
const int SampleSizes[3] = { 512, 256, 128 };
int iSampleSize;

// macros used by the menu and stored in the structure
#define	MODE_PEAK_RMS		1		// bit 0 mask
#define	MODE_PEAK			0
#define	MODE_RMS				1

#define	MODE_VOLTS_DB		2		// bit 1 mask
#define	MODE_VOLTS			0
#define	MODE_DB				2

const char	FFT_MODE[][6] = { "Peak ", "RMS  ", "Volts", "db   " };

// initial values for User Block variables
#define	INI_SAMPLE_FREQUENCY 	45000
#define 	INI_MAX_DISP_FREQUENCY	20000
#define	INI_MIN_DISP_FREQUENCY	100
#define	MAX_SAMPLE_SIZE			512
#define	INI_SAMPLE_SIZE_INDEX	0
#define	INI_MENU_DELAY 			250			// msec
#define	INI_BRIGHTNESS				60				// value
#define	INI_BACKLIGHT				50				// in %
#define	INI_MODE						MODE_RMS|MODE_DB
#define	INI_THRESHOLD				10
#define	INI_DBRANGE					20
#define	INI_REFHT					80				// in %
#define	INI_GAIN_SEL				0				// 0 relative
#define	INI_CAL_FFT_VALUE			12150			// by observation for 512 samples

int	blockexp;								// used by fft function
int	NrFreqs;
float	DC;
int	ADvalues[MAX_SAMPLE_SIZE*2+2];	// array to allow max FFT size
int	*ADPointer;								// pointer to ADvalues - used by ISR
int	ADzero;									// needed by ISR
void	ADinit ( void );
void	ReadAD ( int NbrSamples );


/*******************************************************************/
// There are the simple menu functions and function prototypes
/*******************************************************************/

void fDoFFT ( void );

void fMinDispFreq ( void )
{	UB.MinDispFreq = (int)LCDvalue(UB.MinDispFreq, 5);
}
void fMaxDispFreq ( void )
{	UB.MaxDispFreq = (unsigned)LCDvalue(UB.MaxDispFreq, 5);
}
void fSampleFreq ( void )
{	UB.SampleFreq = (unsigned)LCDvalue(UB.SampleFreq, 5);
}

void fSelGain ( void );
void fDefineGain ( void );
void fFFT_1_Freq ( void );
void fSetADzero ( void );

void fSampleSize128 ( void ) { UB.SampleSizeIndex = 2, iSampleSize = 128; }
void fSampleSize256 ( void ) { UB.SampleSizeIndex = 1, iSampleSize = 256; }
void fSampleSize512 ( void ) { UB.SampleSizeIndex = 0, iSampleSize = 512; }

void fRMS ( void ) { UB.Mode = (UB.Mode&~MODE_PEAK_RMS) | (MODE_RMS); }
void fPeak ( void ) { UB.Mode = (UB.Mode&~MODE_PEAK_RMS) | (MODE_PEAK); }
void fVolts ( void ) { UB.Mode = (UB.Mode&~MODE_VOLTS_DB) | (MODE_VOLTS); }
void fdb ( void ) { UB.Mode = (UB.Mode&~MODE_VOLTS_DB) | (MODE_DB); }
void fdbRange ( void )
{	UB.dbRange = (int)LCDvalue(UB.dbRange, 2); // get and save new db range
}
void fRefHt ( void )
{	UB.RefHt = (int)LCDvalue(UB.RefHt, 2); // get and save new reference ht %
}

void fThreshold ( void )
{	UB.Threshold = (int)LCDvalue(UB.Threshold, 3); // get and save new threshold
}
void fMenuDelay ( void )
{	UB.MenuDelay = (int)LCDvalue(UB.MenuDelay, 4);
}
void fLCDcontrast ( void );
void fLCDbacklight ( void );

void fSettingsRead ( void );
void fSettingsSave ( void );
void fSettingsDisplay ( void );
int	SettingNbr;

/*******************************************************************/
/*******************************************************************/


/*******************************************************************/

/* To use this menu system all the functions and/or their prototypes
		must precede these declarations
Note that the functions CANNOT have any parameters!
See F51553menu.lib for structure details
*/
const _menu FFTMenu[] = {
	{	0,				"Main Menu",			NULL },
	{	1+FUNCTION, "Do FFT",				fDoFFT },
	{	1,			"FFT Setup",				NULL },
	{	2,				"Frequencies",				NULL },
	{	3+FUNCTION,		"Min Disp Freq",			fMinDispFreq },
	{	3+FUNCTION,		"Max Disp Freq",			fMaxDispFreq },
	{	3+FUNCTION,		"Sampling Freq",			fSampleFreq },
	{	2,				"Sample Size",				NULL },
	{	3+FUNCTION,		"512",						fSampleSize512 },
	{	3+FUNCTION,		"256",						fSampleSize256 },
	{	3+FUNCTION,		"128",						fSampleSize128 },
	{	2+FUNCTION, "A/D Gain/Atten",			fSelGain },
	{	1,			"Configure",				NULL },
	{	2,				"RMS vs Peak",				NULL },
	{	3+FUNCTION,		"RMS",						fRMS },
	{	3+FUNCTION,		"Peak",						fPeak },
	{	2,				"Volts vs db",		  		NULL },
	{	3+FUNCTION,		"Volts",						fVolts },
	{	3+FUNCTION,		"db",							fdb },
	{	2+FUNCTION,		"db Range",				fdbRange },
	{	2+FUNCTION,		"Ref Height%",			fRefHt },
	{	1,			"Display",					NULL },
	{	2+FUNCTION,	"Threshold",				fThreshold },
	{	2+FUNCTION,	"Menu Delay",				fMenuDelay },
	{	2+FUNCTION,	"LCD Contrast",			fLCDcontrast },
	{	2+FUNCTION,	"LCD BackLight",			fLCDbacklight },
	{	2+FUNCTION,	"LCD Test",					fTestLCD },
	{	1,			"Settings",					NULL },
	{	2+FUNCTION,	"Read Settings",			fSettingsRead },
	{	2+FUNCTION,	"Save Settings",			fSettingsSave },
	{	2+FUNCTION,	"Display Settings",		fSettingsDisplay },
	{	1, 		"A/D Setup",				NULL },
	{	2+FUNCTION,	"Define Gain/Atten",		fDefineGain },
	{	2+FUNCTION,	"FFT 1-Freq",				fFFT_1_Freq },
	{	2+FUNCTION,	"Set Zero ",				fSetADzero },
	{  0, "", NULL }
};

/*******************************************************************/
/*******************************************************************/
/*******************************************************************/

FFT_DEBUG
void main ( void )
{
auto long	StartTime, ETime;
auto float	f0;
auto int		i, j, k, TestValue;
auto int		FFTsegmentBW;					// convert from interval nbr to freq.
auto int		IndexSize;						// nbr of values in 1 histogram bar
auto int		iCursor;
auto int		MenuItemNbr, FFTsetupItemNbr, FFTsampleSizeNbr;
auto int		ConfigItem;

//	TestIO();

// General Initialization

	LCDfont ( F51553font_2R );				// point to normal font
	F51553_Init();								// initialize the LCD
   ADinit ();									// initialize the A/D
   WrPortI ( RELAY_DDR, &RELAY_SHADOW, RLY_MASK ); // set relay bits to output

   Disable_HW_WDT();							// disable watch dog timer

   LCDclear(0);
	LCDstaticInd ( 2 );						// blink at 0.5 Hz ???????????????
	LCDbrightness ( INI_BRIGHTNESS );

	SettingNbr = 0;							// default setting values
	i=readUserBlock ( &UB, SettingNbr*sizeof(UB), sizeof(UB) ); // read them
   if ( UB.KeyValue != INI_KEY_VALUE ) // here if nothing in User Block yet
   { 	UB.KeyValue = INI_KEY_VALUE;
		UB.SampleFreq = INI_SAMPLE_FREQUENCY;
		UB.MaxDispFreq = INI_MAX_DISP_FREQUENCY;
		UB.MinDispFreq = INI_MIN_DISP_FREQUENCY;
		UB.SampleSizeIndex = INI_SAMPLE_SIZE_INDEX;
		UB.MenuDelay = INI_MENU_DELAY;
		UB.Brightness = INI_BRIGHTNESS;
      UB.Backlight = INI_BACKLIGHT;
		UB.Mode = INI_MODE;
      UB.Threshold = INI_THRESHOLD;
      UB.dbRange = INI_DBRANGE;
      UB.RefHt = INI_REFHT;
      UB.ADzero = 2048;						// default to half scale
      UB.Cal_mv = 500;
		UB.Cal_fft_value[0] = INI_CAL_FFT_VALUE;
		UB.Cal_fft_value[1] = UB.Cal_fft_value[0]/2; // for 256 samples
		UB.Cal_fft_value[2] = UB.Cal_fft_value[0]/4;	// for 128 samples
		UB.GainSel = INI_GAIN_SEL;
      UB.NbrGains = sizeof(GainMask); // get default nbr of gains
		for ( i=0; i<MAXGAINS; i++ )
      {	if ( i<UB.NbrGains )
	      {	UB.dbGain[i] = dbGainValues[i];
				UB.Relays[i] = GainMask[i];
         }
         else
	      {	UB.dbGain[i] = dbGainValues[0];
				UB.Relays[i] = GainMask[0];
         }
      }
      LCDprint ( "No Saved Settings in 0\n\r" );
      LCDprint ( "Select 0\n\r" );
      fSettingsSave();
   }
	iSampleSize = SampleSizes[UB.SampleSizeIndex]; // get the sample size

/*********************************************************/

	LCDclear(0);
	LCDprint ( "   Audio Spectrum\n" );
	LCDprint ( "      Analyzer\n" );
	LCDprint ( "      By K3PTO\n" );
	sprintf ( LCDstring, "    Version %3.1f\n", VERSION );
	LCDprint ( LCDstring );
	LCDprint ( "\nPress Sw 2\n" );
	LCDprint ( "for Main Menu" );

	while ( !(Switches() & SW2_MASK) );	// wait for switch press
	while ( (Switches() & SW2_MASK) );	// wait for switch release
	for ( ;; )									// forever loop
   {
		DoMenu (FFTMenu);
	} // forever loop
} // end main


// enter with data in ADvalues array
FFT_DEBUG
void fDoFFT ( void )
{	auto int		i, j, k;
	auto long	timer, TotalTime, lk;
   auto float	f0, f1;

   auto float	HzPerFFTbucket;
   auto int		HzPerBar;
   auto int		BktsPerBar;					// FFT buckets per display bar
	auto int		Bar0Hz;						// lower frequency of bar 0
   auto int		f0Bucket;					// FFT sample for lowest freq in hist bar
   auto int		f1Bucket;					// FFT sample for highest freq in hist bar
   auto int		fxBucket;					// FFT sample for freqequncy x
	auto int		FFTbucket;					// current FFT bucket being added
	auto int		FFTbarBW;
	auto int		Histogram[LCD_GRAPH_BARS];
   auto int		HistVal, ADval;			// temporary for calculations
   auto int		mode1, mode2;
   auto float	dbref, dbrange, cellsperdb, refcell, cellsperfft;
   auto float	vsensitivity;

// do stuff that only needs to be done once

	SetInterruptRate();						// calculate actual sample frequency
// set the gain relays
	i = UB.Relays[UB.GainSel];
	WrPortI ( RELAY_DR, &RELAY_SHADOW, i ); // select relays

   HzPerFFTbucket = (float)UB.SampleFreq/(float)(iSampleSize*2);
	BktsPerBar = iSampleSize/LCD_GRAPH_BARS; // if using entire freq range

   f0 = (float)(UB.MaxDispFreq-UB.MinDispFreq)/((float)UB.SampleFreq/2.0);
	BktsPerBar = (int)(BktsPerBar*f0 + .9);

   HzPerBar = (int)(BktsPerBar*HzPerFFTbucket + 0.5);
   f0Bucket = (int)(UB.MinDispFreq/HzPerFFTbucket); // MinDispFreq bucket
   f1Bucket = (int)(UB.MaxDispFreq/HzPerFFTbucket); // MaxDispFreq bucket
	Bar0Hz = (int)(f0Bucket * HzPerFFTbucket + 0.5);

// some calculations that do not need to be repeated
	refcell = (float)LCD_GRAPH_HEIGHT*(float)UB.RefHt/100.0;
   												// hist cell for 0db or vref
   cellsperfft = refcell/UB.Cal_fft_value[UB.SampleSizeIndex];
   dbrange = UB.dbRange;
	cellsperdb = (float)LCD_GRAPH_HEIGHT/dbrange;
	mode1 = UB.Mode & MODE_PEAK_RMS; 	// Peak vs RMS
   mode2 = UB.Mode & MODE_VOLTS_DB; 	// Volts vs db
	if ( mode2 == MODE_DB )
   {	dbref = 707.0/(float)UB.Cal_mv;
	   dbref *= (float)UB.Cal_fft_value[UB.SampleSizeIndex]; // reference for db calculation
      vsensitivity = 1.0/cellsperdb;
   }
   else
   {	vsensitivity = (float)UB.Cal_mv/refcell;
   	vsensitivity /= 1000.0;				// convert mv to V.
   }

	LCDclear(0);
	sprintf ( LCDstring, "Mn Dsp Frq: %5d Hz\n", UB.MinDispFreq );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Mx Dsp Frq: %5d Hz\n", UB.MaxDispFreq );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Sample Frq: %5u Hz\n", UB.SampleFreq );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Hz/Bar:     %5d\n", HzPerBar );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Sample Sz:  %5u\n", iSampleSize );
	LCDprint ( LCDstring );
   sprintf ( LCDstring, "Sens: %6.3f", vsensitivity );
	LCDprint ( LCDstring );
	if ( mode2 == MODE_DB )
		LCDprint ( "db/cell\n" );
   else
		LCDprint ( "V/cell\n" );

   LCDprint ("Press Sw1 to start\n\rPress Sw2 to stop" );
 	while ( !(Switches() & SW1_MASK) );	// wait for switch 1 press
 	while ( (Switches() & SW1_MASK) );	// wait for switch 1 release


// check amplitude of input signal
	ReadAD(iSampleSize);
// compute the FFT
	timer = MS_TIMER;
   blockexp = 0;
	fftreal(ADvalues, iSampleSize, &blockexp);
	i = CalcMagnitude();						// calculate the magnitude of the FFT.
	if (i)
   {	LCDclear(0);
   	if ( i<0 ) LCDprint ( "Signal too Low\n\r" );
		else LCDprint ( "Signal too High\n\r" );
	   LCDprint ( "Press Sw1 to Continue\n\r" );
	   LCDprint ( "Press Sw2 to Exit\n\r" );
	 	while ( !(Switches() & (SW1_MASK | SW2_MASK) ) );	// wait for switch press
		if ( SwitchVal & SW2_MASK )
      {	LCDprint ( "Exiting              " );
	 		while ( (Switches() & SW2_MASK) );	// wait for switch 2 release
	      return;
      }
      LCDprint ( "Continuing           " );
		while ( (Switches() & SW1_MASK) );	// wait for switch 1 release
   }

	LCDclear(0);
// print reference indicator
	i = (int)refcell + (64-LCD_GRAPH_HEIGHT); // absolute cell nbr
	LCDhline ( i, LCD_GRAPH_START_BAR - 3, 3 );
	if ( mode2 == MODE_DB )
   {	i = UB.dbGain[UB.GainSel];			// get db gain/atten
		i *= -1;									// 0 db ref = -UB.dbGain[UB.GainSel]
		sprintf ( LCDstring, "\r %3d", i );
   	LCDprint ( LCDstring );
		LCDprint ( "\3 db" );
		// print lower boundary for db
		i = 64-LCD_GRAPH_HEIGHT;			// absolute cell nbr
		LCDhline ( i, LCD_GRAPH_START_BAR - 3, 3 );
      f0 = -1.0*UB.RefHt*UB.dbRange/100.0;  // calculate lower db ref value
      f0 -= UB.dbGain[UB.GainSel];		// adjust for gain
		sprintf ( LCDstring, "\r%4d", (int)f0 );
   	LCDprint ( LCDstring );
   }
	else
	{	f0 = (float)db2vr( UB.dbGain[UB.GainSel] ); // get voltage ratio
		f1 = (float)UB.Cal_mv/1000.0;		// get calibration value in mv
      if ( f0 >= 0.0 ) f1 = f1/f0;		// adjust for gain
      else f1 = f1 * f0;					// adjust for attenuation
		if ( f1 < 0.1 )
      {	sprintf ( LCDstring, "\r%4.2f", f1 );
      }
		else if ( f1 < 1.0 )
      {	sprintf ( LCDstring, "\r%4.1f", f1 );
      }
		else
      {	sprintf ( LCDstring, "\r%4.0", f1 );
      }
   	LCDprint ( LCDstring );
		LCDprint ( "\3 V" );
   }

// print frequency across bottom using small font
// the small font characters use 4 columns including a trailing NULL column
	LCDfont (F51553smNbrs);					// set to use small font
	LCDprint ( "\7" );						// set to page 7

// center of lowest display frequency bar
	i = Bar0Hz + HzPerBar/2;
   sprintf ( LCDstring, "%d^", i );
   i = strlen ( LCDstring );
	LCDsetColumn ( LCD_GRAPH_START_BAR - 4*i + 3 );
	LCDprint ( LCDstring );					// not sure why -3 instead of -2??

// 33%
	j = LCD_GRAPH_BARS*33/100;
	i = Bar0Hz + (j * HzPerBar) + HzPerBar/2;
   sprintf ( LCDstring, "%d^", i );
   i = strlen ( LCDstring );
	LCDsetColumn ( LCD_GRAPH_START_BAR + j - 4*i + 3 );
	LCDprint ( LCDstring );					// not sure why -3 instead of -2??

// 66%
	j = LCD_GRAPH_BARS*66/100;
	i = Bar0Hz + (j * HzPerBar) + HzPerBar/2;
   sprintf ( LCDstring, "%d^", i );
   i = strlen ( LCDstring );
	LCDsetColumn ( LCD_GRAPH_START_BAR + j - 4*i + 3 );
	LCDprint ( LCDstring );					// not sure why -3 instead of -2??

// center of highest display frequency less 1 bar
	j = LCD_GRAPH_BARS - 2;					// relative bar
	i = Bar0Hz + (j * HzPerBar) + HzPerBar/2;
   sprintf ( LCDstring, "%d^", i );
   i = strlen ( LCDstring );
	LCDsetColumn ( LCD_GRAPH_START_BAR + j - 4*i + 3 );
	LCDprint ( LCDstring );					// not sure why -3 instead of -2??

	LCDfont ( F51553font_2R );				// back to normal font

//	LCDvline ( 56, LCD_GRAPH_START_BAR );
//	LCDvline ( 56, LCD_GRAPH_LAST_BAR-1 );
	;
	while (1)		// start FFT loop	*******************************************
   {

	TotalTime = MS_TIMER;
#ifdef USE_DUMMY_DATA
	DummyData ();
#else
	ReadAD(iSampleSize);
#endif

// compute the FFT
	timer = MS_TIMER;
   blockexp = 0;
	fftreal(ADvalues, iSampleSize, &blockexp);

#if DEBUG_FFT_DISPLAY & 1
	timer = MS_TIMER-timer;
   printf("\n%4d-point fftreal()           time %ld ms\n",
   	iSampleSize, timer);
	DumpADvalues( "after fftreal()" );
#endif

	i = CalcMagnitude();						// calculate the magnitude of the FFT.
	if (i)
   {
		goto DoFFT_0;
   }
// Analysis Frequency = K * FFT array index
// K = FrequencyFactor = Sample Frequency / FFTsize*2
//	UB.Cal_mv = (int)(VRMSavg*1000.0);	// save calibration voltage
//	UB.Cal_fft_value = Vmax;				// and FFT value

// calculate histogram values - pick RMS or peak value within the bar
	timer = MS_TIMER;
   for ( i=0; i<LCD_GRAPH_BARS; i++ )	// for each bar
	{	HistVal = 0;
   	f0 = 0.0;								// init for RMS calculation
		j = f0Bucket + (i*BktsPerBar);	// starting bucket for this bar
      for ( k=j; k<j+BktsPerBar; k++ )
      {	ADval = ADvalues[k];
      	if ( mode1 == MODE_PEAK )		// if mode is Peak
         {	if ( ADval > HistVal )
         		HistVal = ADval;
			}
			if ( mode1 == MODE_RMS )		// if mode is RMS
         {	f1 = (float)ADval;
				f0 += f1*f1;					// sum of squares
         }
		}

// For the fft buckets making up this histogram bar:
//		Peak mode: HistVal = maximum fft value
//		RMS mode: f0 = sum of squares

		if ( mode1 == MODE_RMS )			// if mode is RMS
	     	f0 = sqrt(f0);						// calc RMS value
		else
     		f0 = HistVal;						// convert fft value to float


		if ( mode2 == MODE_DB )				// if mode is db
		{	if ( f0 < 1.0 ) f0 = 1.0;		// do not allow 0.0
			f0 = 20.0*log10(f0/dbref);		// calc db
			f0 = f0*cellsperdb + refcell;	// absolute cell nbr
      }
		else										// mode is volts
      {	f0 = f0*cellsperfft;
      }
      j = (int)f0;							// convert to integer
		if ( j < 0 ) j = 0;					// negative values do not display
		Histogram[i] = j;						// save for graph

      ;	// for breakpoint
   } // for each bar

#if DEBUG_FFT_DISPLAY & 8
	timer = MS_TIMER-timer;
	printf ( "\nFFT column values     time = %ld\n\r", timer );
   for ( i=0; i<LCD_GRAPH_BARS; i+=4  )
   	printf( "%4d  %6d  %6d  %6d  %6d\n",
      	i, Histogram[i], Histogram[i+1], Histogram[i+2], Histogram[i+3] );
#endif

	timer = MS_TIMER;
	for ( i=0; i<LCD_GRAPH_BARS; i++ )
   {	LCDvline ( Histogram[i], LCD_GRAPH_START_BAR + i );
   }
	timer = MS_TIMER-timer;

   TotalTime = MS_TIMER - TotalTime;	// total per analysis cycle
	f0 = 1000.0/(float)TotalTime;			// calculate update frequency
//   sprintf ( LCDstring, "%4.1f", f0 );
//   LCDprint ("\0");
//   LCDprint ( LCDstring );
   ;												// for breakpoint
	DoFFT_0:
 	if ( (Switches() & SW2_MASK) ) break; // if switch 2 press
   ;

	} // end FFT loop	**********************************************************
	LCDprint ( "\7Exiting              " );
 	while ( (Switches() & SW2_MASK) );	// wait for switch 2 release

	; // for break point
} // DoFFT

/* Calculate actual SampleFreq based on UB.SampleFreq
The interrupt is generated by Timer A7 driven by Timer A1.  This gives a total
maximum divisor value of 256*256.  Timer A1 is driven by PClk/2 = CPUfreq/2.
The interrupt frequency becomes the sample frequency since the ISR takes one
sample of the input signal.  This frequency should be at least twice the highest
input frequency to be displayed.  Therefore, the divisor can
be calculated as follows:
	Divisor = CPUfreq/2/UB.Samplfreq
*/
FFT_DEBUG
void SetInterruptRate ( void )
{	auto float	fCPUfreq, fDivisor;
	int	iDivisor, TAT1divisor, TAT7divisor, TAT1temp, TAT7temp, DivisorTemp;
   unsigned	SmpFreq;
   int	i;

// adjust sample freq to be next larger multiple of SampleSize
	SmpFreq = UB.SampleFreq/iSampleSize + 1;
   SmpFreq *= iSampleSize;

// calculate the total divisor value
	fCPUfreq = (float)freq_divider * 614400.0;
	fDivisor = (fCPUfreq/2.0/SmpFreq);
   iDivisor = (int)(fDivisor+0.5);

// find the values of TAT1divisor and TAT7divisor which yield the divisor value
// closest to iDivisor
	TAT1divisor = TAT7divisor = (int)sqrt(fDivisor); // init values
   DivisorTemp = TAT1divisor*TAT7divisor;
	for ( TAT1temp = iDivisor/256; TAT1temp<255; TAT1temp++ )
   {	TAT7temp = iDivisor/TAT1temp;
   	if ( TAT7temp <= 256 )
      { 	i = TAT1temp * TAT7temp;			// calculate total divisor
   		if (	abs(i-iDivisor) < abs(i-DivisorTemp) )
      	{	TAT1divisor = TAT1temp;
				TAT7divisor = TAT7temp;
			   DivisorTemp = TAT1divisor*TAT7divisor;
   	   }
      	if (i-DivisorTemp == 0) break;
      }
	}
//	UB.SampleFreq = (unsigned)((fCPUfreq/2.0/(float)DivisorTemp)+0.5);

// Set up the timer interrupt interval but do not enable the interrupt
   WrPortI ( TACR, &TACRShadow, 0x80 ); // TA7 clocked by TA1, IPL = 0
   WrPortI ( TAT1R, &TAT1RShadow, TAT1divisor-1);
   WrPortI ( TAT7R, &TAT7RShadow, TAT7divisor-1);
	SetVectIntern ( 0x0A, TimerISR );	// point to ISR
   WrPortI ( TACSR, &TACSRShadow, 0x81 ); // TA7 int enable, timer A enabled

   return;
} // SetInterruptRate

/* This function allows the operator to vary both the V5 voltage regulator and
the electronic Volume (contrast) control.  Bits 3-8 are the Volume, bits 0-2 are
the V5 value.  They are handled as a single integer so the V5 will go through 8
steps before the Volume value is incremented or decremented.  If you go
too far you will not be able to see the characters on the display.

The contrast seems to be more sensitive to the Volume value than V5.
*/
FFT_DEBUG
void fLCDcontrast ( void )
{	int CurrentValue,i,j;

	LCDprint ("\nPress Sw1 to accept\n" );
	LCDprint ("Press Sw2 to increase\n" );
	LCDprint ("Press Sw3 to decrease\n" );

	CurrentValue = ((60<<3) + 6);			// same as init function
	while (1)
   {
		i = CurrentValue & 0x07;			// V5 resistor ratio ( LS 3 bits )
      j = CurrentValue >> 3;				// volume set value ( MS 6 bits )
		CSenable();
	   TxInst ( i + 0x20 );					// offset of 0x20
   	TxInst ( 0x81 );						// volume set command
	   TxInst ( j );
		CSdisable();

		sprintf ( LCDstring, "V5 = %1X   Vol = %02X\r", i, j );
		LCDprint ( LCDstring );
		while ( !Switches() );				// wait for sw press
	   switch ( Switches() )
   	{
      case SW_SELECT:
		 	while ( (Switches() & SW_SELECT) );	// wait for sw release
			return;
	   	break;

		case SW_ADV:
			CurrentValue += 1;
	   	break;

		case SW_DEC:
			CurrentValue -= 1;
   		break;

      default:
	   } // switch

   	if ( CurrentValue <= 1 ) CurrentValue = 1;
		if ( CurrentValue >= 0x1FF ) CurrentValue = 0x1FF;
      MsDelay(UB.MenuDelay);				// keep from going too fast
   } // while
} // fLCDcontrast

FFT_DEBUG
void fLCDbacklight ( void )
{	int i, CurrentValue;

	LCDprint ("\nPress Sw1 to accept\n" );
	LCDprint ("Press Sw2 to increase\n" );
	LCDprint ("Press Sw3 to decrease\n" );

	CurrentValue = UB.Backlight;
	while (1)
   {
		sprintf ( LCDstring, "Value = %2d\r", UB.Backlight );
		LCDprint ( LCDstring );
	   switch ( Switches() )
   	{
      case SW_SELECT:
		 	while ( (Switches() & SW_SELECT) );	// wait for sw release
			return;
	   	break;

		case SW_ADV:
			UB.Backlight += 1;
	   	break;

		case SW_DEC:
			UB.Backlight -= 1;
   		break;

      default:
	   } // switch

		if ( CurrentValue != UB.Backlight )
      {	if ( UB.Backlight <= 1 ) UB.Backlight = 1;
			if ( UB.Backlight >= 99 ) UB.Backlight = 99;
         i = 256*UB.Backlight/100;
			WrPortI ( PWM0R+(2*LCD_PWM_CH), NULL, i);
         CurrentValue = UB.Backlight;
         MsDelay(UB.MenuDelay);			// keep from going too fast
      }
   } // while
} // fLCDbacklight

// read the settings saved in the UserBlock
FFT_DEBUG
void	fSettingsRead ( void )
{
	SettingNbr = (int)LCDvalue(SettingNbr, 1);	// get and save setting nbr
	readUserBlock ( &UB, SettingNbr*sizeof(UB), sizeof(UB) );
} // fSettingsRead


// save the settings to the UserBlock
FFT_DEBUG
void	fSettingsSave ( void )
{	int i;
	SettingNbr = (int)LCDvalue(SettingNbr, 1);	// get and save setting nbr
	i = writeUserBlock( SettingNbr*sizeof(UB), &UB, sizeof(UB) );
} // fSettingsSave


FFT_DEBUG
void	fWaitSw2 ( int Flag )
{	if ( !Flag ) LCDprint ( "\7Press Sw2 to Cont." );
	else LCDprint ( "\7Press Sw2 to Exit" );
 	while ( !(Switches() & SW_ADV) );	// wait for sw press
 	while ( (Switches() & SW_ADV) );		// wait for sw release
}

FFT_DEBUG
void	fSettingsDisplay ( void )
{	auto int i,j,k, mode;
	auto float ftemp;
   auto char tempstr[10];

// NOTE: only six lines of data can be displayed per page

	LCDclear(0);
	LCDprint ( "  FFT Settings\n" );
	sprintf ( LCDstring, "Sample Frq: %5u\n", UB.SampleFreq );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Mx Dsp Frq: %5d\n", UB.MaxDispFreq );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Mn Dsp Frq: %5d\n", UB.MinDispFreq );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Sample Size:%5d\n", iSampleSize );
	LCDprint ( LCDstring );

	fWaitSw2(0);

	LCDclear(0);
	LCDprint ( "  A/D Settings\n" );
	i = UB.Mode & MODE_PEAK_RMS; 			// Peak vs RMS
	mode = UB.Mode & MODE_VOLTS_DB; 		// Volts vs db
	sprintf ( LCDstring, "A/D Mode: %s %s\n", FFT_MODE[i/MODE_PEAK_RMS],
   	FFT_MODE[mode/MODE_VOLTS_DB+2] );
	LCDprint ( LCDstring );

	sprintf ( LCDstring, "Threshold:%4d\n", UB.Threshold );
	LCDprint ( LCDstring );
	ftemp = (float)UB.ADzero * 5.0 / 4096.0;
	sprintf ( LCDstring, "A/D zero: %4d=%5.3fv\n", UB.ADzero, ftemp );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Cal mv:   %4d\n", UB.Cal_mv );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Cal FFT value:%5d\n", UB.Cal_fft_value[UB.SampleSizeIndex] );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "db Range: %4d\n", UB.dbRange );
	LCDprint ( LCDstring );

	fWaitSw2(0);

	fDispGains(0);								// display defined gains
	sprintf ( LCDstring, "Gain Selected = #%2d\n", UB.GainSel );
	LCDprint ( LCDstring );

	fWaitSw2(0);

	LCDclear(0);
	LCDprint ( "  Display Settings\n" );
	sprintf ( LCDstring, "Ref Ht:    %4d%%\n", UB.RefHt );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Menu Delay:%4dms\n", UB.MenuDelay );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Contrast:  %4d\n", UB.Brightness );
	LCDprint ( LCDstring );
	sprintf ( LCDstring, "Backlight: %4d%%\n", UB.Backlight );
	LCDprint ( LCDstring );

	fWaitSw2(1);

} // fSettingsDisplay



/********************************************************************
								A/D Routines
********************************************************************/

FFT_DEBUG
void fDispGains ( int Flag )
{	auto int i, j, k, mode, count;

	LCDclear (0);
	LCDprint ( "Gain/Atten Values:\n" );
   mode = UB.Mode & MODE_VOLTS_DB;
   count = MAXGAINS;							// assume do all possible gains
   if ( !Flag ) count = UB.NbrGains;	// just do defined gains
	for (i=0; i<count; i++ )
   {	j = i%4;									// 4 gains per row
		k =  UB.dbGain[i];					// get gain value
		if ( mode == MODE_VOLTS ) k = db2vr(k); // convert to ratio
		if ( !j )
      {	LCDsetCharPos(0);
			sprintf ( LCDstring, "%2d:%3d", i, k );
		}
      else
      {	LCDsetCharPos( (j-1)*4 + 8);	// allow 4 characters per value
			sprintf ( LCDstring, "%3d", k );
      }
		LCDprint ( LCDstring );
      if ( j == 3 ) LCDprint("\n\r");	// next line
   }
   if ( j != 3 ) LCDprint("\n\r");		// next line - if needed
} // fDispGains


FFT_DEBUG
void fSelGain ( void )
{
	fDispGains(0);								// display defined gains
	UB.GainSel = (int)LCDvalue(UB.GainSel, 2|LCDVALUE_NONL);
} // fSelGain


FFT_DEBUG
void fDefineGain ( void )
{	auto int i, j, k, mode;
	auto float ftemp;
   auto int GainNbr;

	fDispGains(1);								// display all gains possible
	GainNbr = (int)LCDvalue( 0L, 2|LCDVALUE_NONL);	// get gain nbr to modify
	if ( GainNbr<0  ||  GainNbr>MAXGAINS-1 ) GainNbr = 0;	// force valid value
   if (GainNbr >= UB.NbrGains) UB.NbrGains = GainNbr+1; // if additional gain

	LCDclear (0);								// clear LCD starting with page 0
   mode = UB.Mode & MODE_VOLTS_DB;
	if ( mode == MODE_DB )					// if db
   {	sprintf ( LCDstring, "Enter value %2d in db", GainNbr );
		LCDprint ( LCDstring );
		UB.dbGain[GainNbr] = (int)LCDvalue(UB.dbGain[GainNbr], LCDVALUE_NEG|3);
   }
	else		// if voltage
   {	sprintf ( LCDstring, "Enter value %2d as\n\r   a voltage ratio", GainNbr );
		LCDprint ( LCDstring );
		i = db2vr(UB.dbGain[GainNbr]);	// convert db to vr
		i = (int)LCDvalue(i, LCDVALUE_NEG|4); // get the new value
      UB.dbGain[GainNbr] = vr2db(i);	// convert to int and save
   }
   LCDprint ( "\rRelay pattern:       " );
	UB.Relays[GainNbr] = (int)LCDvalue( UB.Relays[GainNbr], 2|LCDVALUE_HEX );
} // fDefineGain

// convert a voltage ratio to db
// NOTE: a negative db value will be converted to a negative integer
FFT_DEBUG
int vr2db ( int VR )
{	auto float fvr;

	fvr = VR;									// convert to float
   if ( VR < 0 ) fvr *= -1.0;				// make positive
	fvr = 20.0 * log10 ( fvr ) + 0.5;	// convert ratio to db
	if ( VR < 0 ) fvr *= -1.0;				// restore sign
	VR = (int)fvr;
   return (VR);
}

// convert db to a voltage ratio
FFT_DEBUG
int db2vr ( int db )
{	auto int k;
	auto float ftemp;

	k = 1;										// assume gain
   if ( db<0 ) k = -1;						// show attenuation
  	ftemp = pow10 ( (float)(db*k)/20.0 ); // calc ratio
   k = k * (int)ftemp;						// convert to int and adj for atten
	return (k);
}



/* ADinit must be aware of the serial port settings used by the LCD and must
be executed AFTER the F51553_Init function.
*/
FFT_DEBUG
void	ADinit ( void )
{
	BitWrPortI ( AD_DR, &AD_SHADOW, 1, AD_CS_BIT );// insure A/D is not enabled
	BitWrPortI ( AD_DDR, &AD_DDR_SHADOW, 1, AD_CS_BIT );// make /CS bit an output
} // ADinit


/* Read the A/D convertor
		Take NbrSamples reading and store into ADvalues
		The Timer ISR is set to IPL 2 to insure a consistant sample interval
*/

FFT_DEBUG
void ReadAD ( int NbrSamples )
{	auto int* ADPointer1, i;
	auto long timer;

#if DEBUG_FFT_DISPLAY & 0x10
	timer = MS_TIMER;
	memset ( &ADvalues[0], 0, sizeof(ADvalues) );
#endif
	ADzero = UB.ADzero;						// needed by A/D ISR
	ADPointer = &ADvalues[0];				// point to array
	ADPointer1 = &ADvalues[NbrSamples*2]; // point to last element of array
   WrPortI ( TACR, &TACRShadow, 0x81 ); // TA7 clocked by TA1, IPL = 2
	while ( ADPointer < ADPointer1 ); 	// wait for all samples
   WrPortI ( TACR, &TACRShadow, 0x80 ); // TA7 clocked by TA1, IPL = 0

#if DEBUG_FFT_DISPLAY & 0x10
	timer = MS_TIMER-timer;
   printf("Raw A/D data      time = %ld\n\r", timer);
	printf ( "ADzero = %4d    NbrSamples = %3d\n\r", ADzero, NbrSamples );
	for ( i=0; i<2*NbrSamples; i+=4 )
   	printf ( "%4d %6d %6d %6d %6d\n",
      	i, ADvalues[i], ADvalues[i+1], ADvalues[i+2], ADvalues[i+3] );
#endif
	;
// now process the readings
} // ReadAD

/* If the SPI clock is 2Mbps, this ISR will take about 16usec with a 29.5MHz
clock and about 12usec witha 59MHz clock.

This ISR should be executed at IPL = 2 or 3 so that it connot be interrupted.
This is required so the sample rate is consistant.

This ISR should be enabled with ADPointer = address to put the readings.  It
must be disabled when the program detects a full sample has been accumulated.

This code breaks one of the main "rules" for ISRs:
					An ISR should not have any loops.
This is done here only because of the dedicated nature of the program.

The format of the data from the ADS7816 is:
	1st byte: X  11 10 9  8  7  6  5
   2nd byte: 4  3  2  1  0  X  X  X
The voltage is always positive from 0x0000 to 0x0FFF ( 0 to +VRef ).  The data
handling algorithm in the ISR shifts the data by two bits to align it properly.
It then subtracts a calibration constant (ADzero) from the value to force it to
center around 0.

The variable ADzero is used insted of UB.ADzero beacuse Dynamic C assembly
language does not have a construct for accessing structure elements.
*/

#asm
TimerISR::
			push	af								; 10
	ioi	ld		a, (TACSR)					; 11 ack the interrupt and clear it
			ADCS_ON								; 23 enable the A/D /CS
   ioi   ld    a, (SBAR)               ; 11 start an SPI read
														; 55
; The following three instructions take essentially 0 time since we need
; to wait for the data byte.
; With a 2MBaud clock each bit takes about 15/30 CPU clocks.  This yields a total
; byte time of about 240 CPU clocks
         push	ix								; 10
			push	de								; 10 used for data manipulation
			push	HL								; 10 used for data manipulation
; wait for MS byte
TimerISR_1:
	ioi	ld		a, (SBSR)					; 11 read status register
			rla									;  4 bit 7 -> C = 1 if byte available
         jr		nc, TimerISR_1				;  5 jump if byte not ready yet
														; 120/240
;TimerISR_1 loop

	ioi	ld		a, (SBAR)					; 11 read the byte and start next one
			ld		H, a							;  4 save MS byte
														; 35
; The following two instructions take essentially 0 time since we need
; to wait for the data byte.
			ld		ix, (ADPointer)			; 13 get address to put the data
			ld		de, (ADzero)				; 13 AD value with no input
; wait for second byte
TimerISR_2:
	ioi	ld		a, (SBSR)					; 11 read status register
			rla									;  4 bit 7 -> C = 1 if byte available
         jr		nc, TimerISR_2				;  5 jump if byte not ready yet
														; 120/240
;TimerISR_2 loop

	ioi	ld		a, (SBDR)					; 11 read the byte
			ld		L, a							;  4 save LS byte - make 16 bit integer
         ADCS_OFF								; 23 disable A/D /CS (uses a reg)
														; 58
; create and process the 16 bit integer value
			rr		HL								;  2 adjust for
         rr		HL								;  2 		shifted bits
         ld		a, H							;  4 get MS byte
         and	0x0F							;  4 keep only 4 MS bits (also C = 0)
         ld		H, a							;  4 put back into integer value
			sbc	HL, de						;  4 force symmetry around zero
;														; 20

			ld		(ix), HL						; 11 save the reading
         inc	ix								;  4 point to next reading address
         inc	ix								;  4
			ld		(ADPointer), ix			; 15 save pointer for next reading
			pop	HL								;  7
			pop	de								;  7
			pop	ix								;  9
   		pop	af								;  7
         ipres									;  4
         ret									;  8
														; 76
#endasm
// TimerISR

/* This function is basically a test of the A/D and the FFT algorithm.  It can
be used to calibrate the amplitude response of the system.  It assumes that the
input is a single frequency sine wave.  Here is an outline:
	Take SampleSize*2 samples
   Calculate the RMS value by finding the min and max values.
	Do the FFT analysis
   Find the "bucket" with the highest amplitude result.  Use this bucket and the
   	two on either side of it.
   Display the RMS value, bucket nbr and amplitude

These two values must be stored into UserBlock: UB.Cal_mv, UB.Cal_fft_value
*/
FFT_DEBUG
void fFFT_1_Freq ()
{	auto float	HzPerFFTbucket;
   auto int		BktsPerBar;					// FFT buckets per display bar
	auto int		i, MaxFFTindex;
   auto int		Vmin, Vmax;
   auto float	VRMS, VRMSavg;
   auto int		FFTfreq;
   auto int		Iteration;
   auto float	FFTampavg, A1,A2,A3;
   auto int		CurrentSampleSizeIndex;

	SetInterruptRate();						// set sample rate
	i = UB.Relays[UB.GainSel];				// get relay control value
	WrPortI ( RELAY_DR, &RELAY_SHADOW, i ); // activate relays

	LCDprint ( "\nVoltage should be   \n\r" );
	if ( (UB.Mode & MODE_VOLTS_DB) == MODE_VOLTS )
	   sprintf ( LCDstring, " close to 1.0v" );
   else
	   sprintf ( LCDstring, " close to 0.707v" );
   LCDprint ( LCDstring );
   LCDprint ( "\7Press Sw 2 to start" );
 	while ( !(Switches() & SW2_MASK) );	// wait for switch press
 	while ( (Switches() & SW2_MASK) );	// wait for switch release
	LCDprint ( "\5VRMS   Freq  FFTampl\n\r" );

	for ( CurrentSampleSizeIndex=0; CurrentSampleSizeIndex<=2; CurrentSampleSizeIndex++ )
   {
		sprintf ( LCDstring, "\4Sample Size = %d", SampleSizes[CurrentSampleSizeIndex] );
      if ( CurrentSampleSizeIndex != 2 ) LCDprint ( "\7Press Sw2 to Advance" );
      else LCDprint ( "\7Press Sw2 to Exit" );
      LCDprint ( LCDstring );
		iSampleSize = SampleSizes[CurrentSampleSizeIndex];
   	HzPerFFTbucket = (float)(UB.SampleFreq/2.0)/(float)iSampleSize;
		BktsPerBar =(int)( ( (UB.MaxDispFreq - UB.MinDispFreq)/HzPerFFTbucket )
   				/ LCD_GRAPH_BARS );
		if ( BktsPerBar < 1 ) BktsPerBar = 1;

   	Iteration = 0;
	   FFTampavg = 0.0;
   	VRMSavg = 0.0;
		while (1)
   	{
			ReadAD(iSampleSize);	// get values into ADvalues
// calculate the RMS value
			Vmin = 32767;
      	Vmax = 0;
	      for ( i=0; i<iSampleSize*2; i++ )
   	   {	if ( ADvalues[i] > Vmax ) Vmax = ADvalues[i];
      		if ( ADvalues[i] < Vmin ) Vmin = ADvalues[i];
	      }
			VRMS = (float)(Vmax-Vmin)/2.828;	// Pk-pk/2*sqrt(2)
			VRMS = VRMS*5.0/4096.0;				// adjust for 5V and 12 bits A/D
			VRMSavg += VRMS;

		   blockexp = 0;							// required by fftreal
			fftreal(ADvalues, iSampleSize, &blockexp);
			CalcMagnitude();						// calculate the magnitude of the FFT.
   	  	MaxFFTindex = 0; 						// index into ADvalues for max value
     		for ( i=1; i<iSampleSize; i++ )	// do not use sample 0 ??????????
			{	if ( ADvalues[i] > ADvalues[MaxFFTindex] ) MaxFFTindex = i;
   	  	}
		// calculate the RMS value of the 3 adjacent buckets
			A1 = (float)ADvalues[MaxFFTindex-1];
			A2 = (float)ADvalues[MaxFFTindex];
			A3 = (float)ADvalues[MaxFFTindex+1];
			A1 *= A1;
			A2 *= A2;
			A3 *= A3;
			FFTampavg += sqrt(A1+A2+A3);

	      Iteration += 1;
   	   if ( Iteration == 10 )
      	{	FFTampavg /= 10.0;				// avg of 10 readings
      		Vmax = (int)FFTampavg;			// convert to int
				FFTfreq = (int)(HzPerFFTbucket*MaxFFTindex);
   	      VRMSavg /= 10.0;
				sprintf ( LCDstring, "\6%5.3f %5d  %5d",
      					VRMSavg, FFTfreq, Vmax );
		      LCDprint ( LCDstring );
			 	if ( (Switches() & SW2_MASK) ) break; // exit if switch 2 press
			   Iteration = 0;						// reset for
		   	FFTampavg = 0.0;					//		next set
			   VRMSavg = 0.0;
   	   }
		} //  while (1)
		UB.Cal_mv = (int)(VRMSavg*1000.0);	// save calibration voltage
		UB.Cal_fft_value[CurrentSampleSizeIndex] = (int)FFTampavg;	// and FFT value
      if ( CurrentSampleSizeIndex != 2 ) LCDprint ( "\7Advancing            " );
      else LCDprint ( "\7Exiting              " );
	 	while ( (Switches() & SW2_MASK) );	// wait for switch 2 release
	} // for each sample size

	iSampleSize = SampleSizes[UB.SampleSizeIndex]; // restore sample size
} // fFFT_1_Freq


/* This function allows you to switch the gain blocks
in and out.  This allows you to adjust the zero adjust potentiometers
so the outputs are the same for all gains.
The result is stored in the array for the UserBlock in UB.ADzero.

For Gain Select = 0db the actual voltage is displayed.
When the gain blocks are enabled the voltage is displayed relative to the
Gain 0 value.  The potentiometers should be adjusted for 0.
*/
FFT_DEBUG
void fSetADzero ( void )
{	float volts;
	long ReadingSum;
   int i, NbrReadings, ADref, CalIndex;

	SetInterruptRate();						// set sample rate
	UB.ADzero = 0;								// init to no offset
	NbrReadings = 128;
   CalIndex = 0;								// point to reference setting

	LCDprint ( "\1Disconnect Input Sig " );
	LCDprint ( "\6Sw3 to change gain   " );
	LCDprint ( "\7Press Sw2 to Exit    \r" );
   LCDsetPage (3);
	while (1)
   {
		WrPortI ( RELAY_DR, &RELAY_SHADOW, CalRelays[CalIndex] );
      MsDelay (100);
		ReadAD ( NbrReadings );
      ReadingSum = 0L;
      for ( i=0; i<NbrReadings; i++ )
      	ReadingSum += (long)ADvalues[i];
      ReadingSum /= (long)NbrReadings; // get average
   	if ( CalIndex == 0 )
		{	ADref = (int)ReadingSum & 0xFFF;
      }
      else
		{ 	ReadingSum -= ADref;
		}
		volts = (float)ReadingSum * 5.0 / 4096.0;
      sprintf ( LCDstring, "%s %6.3fV    \r", CalPots[CalIndex], volts );
		LCDprint ( LCDstring );
	 	if ( (Switches() & SW2_MASK) ) break; // if switch 2 press
	 	if ( (Switches() & SW3_MASK) )	// if switch 3 press
      { 	while ( (Switches() & SW3_MASK) );// wait for switch 3 release
			CalIndex += 1;
         if ( CalIndex >= sizeof(CalRelays) ) CalIndex = 0;
      }
	}
	UB.ADzero = ADref;						// store the '0' value
	LCDprint ( "\7Exiting              " );
 	while ( (Switches() & SW2_MASK) );	// wait for switch 2 release
   ; // for breakpoint
} // fSetADzero


/* Calculate the magnitude of the real and imaginary components
	and convert back to integer
	The real part is in the even numbered elements, the imaginary part is in the
   next higher numbered element.
   The result is put back into the array starting with ADvalues[0]

These variables are global due to the assembly language code:
it is easier to access global variables than locals

Return value:
					-1 = input signal too small
                0 = success
                1 = input signal too large
*/
long L0, L1;
int *p0, *p1;
const int Pwr2[] = { 1, 2, 4, 8, 16, 32, 64, 128, 256 };

FFT_DEBUG
int CalcMagnitude ( void )
{	auto float f0, f1;
	auto int i;
   auto long Div2val;						// divide by 2 count
   auto long temp;
   auto int *p3;
	auto long timer;
   auto int retval;

	timer = MS_TIMER;
   retval = 0;									// assume success
	if ( blockexp > 5 )
   {	retval = 1;					 			// show signal too large
   	blockexp = 5;
   }
   if ( blockexp < 0 )
   {	retval = -1;							// show signal too small
		blockexp = 0;
   }
	Div2val = Pwr2[5-blockexp];			// divide FFT value by 128/2^blockexp
	p0 = p3 = &ADvalues[0];
	p1 = &ADvalues[1];
   for ( i=0; i<iSampleSize; i++ )
	{
//   	L0 = *p0;								// get real part as long
//   	L1 = *p1;								// get imaginary part as long
//   	L0 = (L0)^2 + (L1)^2;				// sum of squares as long
//		ADvalues[i] = sqrt(L0)
#asm
	// get L0
   	ld		HL, (p0)
      ld		HL, (HL)
	// square L0
;   	ld		HL, (L0)
		ld		B, H
      ld		C, L
      ld		D, H
      ld		E, L
      mul					; HL:BC = L0*L0
      ld		D, B
      ld		E, C			; HL:DE = L0*L0
		exx					; save in alternates - HL':DE' = L0*L0
	// get L1
   	ld		HL, (p1)
      ld		HL, (HL)
	// square L1
;   	ld		HL, (L1)
		ld		B, H
      ld		C, L
      ld		D, H
      ld		E, L
      mul					; HL:BC = L1*L1
      ld		D, H			; save MSword
      ld		E, L
	// add the squares together
		ex		DE', HL		; retrieve LSword of L0*L0 into HL
      add	HL, BC		; add the LSwords
      ld		(L0), HL		; store LSword
      ld		H, D			; retrieve MSword
      ld		L, E
	altd ex	DE, HL		;DE <-> HL': retrieve MSword of L0*L0 into DE
		adc	HL, DE		; HL = sum of MSwords + carry
      ld		(L0+2), HL	; store MSword
#endasm
		f0 = (float) L0;						// convert to float
		f0 = sqrt( f0 );						// magnitude
		temp = (long)f0;						// convert to long
      temp /= Div2val;						// scale it
      *p3++ = (int)temp;					// back to integer and save
   	p0 += 2;									// point to next real value
      p1 += 2;									// point to next imaginary value
   }
#if DEBUG_FFT_DISPLAY & 2
	timer = MS_TIMER-timer;
	printf ( "\nMagnitude of FFT values     time = %ld\n\r", timer );
   for (i = 0; i < iSampleSize; i+=4)
   	printf("%4d  %6d  %6d  %6d  %6d\n",
   		i, ADvalues[i], ADvalues[i+1], ADvalues[i+2], ADvalues[i+3]);
#endif
	return (retval);							// 0 = success
} // CalcMagnitude


/* Fill the LCD in order to test all positions of the display.
*/
FFT_DEBUG
void fTestLCD ( void )
{	char TestChar[2];
	int i, page;

	TestChar[0] = 0x30;							// start with '0'
   TestChar[1] = 0;
	LCDsetPage ( 0 );
	for (page=1; page<=7; page++ )		// use each page
	{	LCDprint ( "\n\r" );
		for ( i=0; i<=20; i++ )
      {	LCDprint ( TestChar );
      	TestChar[0] += 1;
         if ( TestChar[0] >= 0x7E ) TestChar[0] = 0x30;
		} // next char
   } // next page
   MsDelay ( 1000 );
	LCDprint ( "\7Press Sw2 to Exit     " );
 	while ( !(Switches() & SW2_MASK) );	// wait for switch 2 press
 	while ( (Switches() & SW2_MASK) );	// wait for switch 2 release
}



FFT_DEBUG
void DumpADvalues( char *title ) {
    auto int k;
    printf("ADvalues[n] %s\n", title);
    for (k = 0; k < DEBUG_PRINT_ADvalues; k+=2)
        printf("%4d  %6d %6d   %6d %6d\n",
	        k, ADvalues[k*2], ADvalues[k*2 + 1],
           ADvalues[k*2 + 2], ADvalues[k*2 + 3]);
    printf("blockexp is %d\n\n", blockexp);
} // DumpADvalues


#ifdef USE_DUMMY_DATA	// ****************************************************

// the following are used to create dummy data for testing
#define     max(a,b)    ( a > b ? a : b )
#define		MAXFREQS	4

// loads ADvalues[] with dummy data values
FFT_DEBUG
void DummyData ( void )
{
auto int i;
auto int k;
auto float Vmax, fscale, dc, ftemp;
auto long j;
auto float	FreqMult[MAXFREQS];
auto float	Freq[MAXFREQS];
auto int		Ampl[MAXFREQS];

// set up dummy data
//  First construct a real-valued floating-point array comprising a dc value and
//  up to MAXFREQS sinusoids with configurable amplitude, frequency, and phase.
//  Round the floating values to integers, scaling if necessary to fit in a
//  16-bit word, store them in the integer array x[], and print x[].  Then
//  transform x[] via fftreal() and print the results.  Finally inverse
//  transform x[] via fftrealinv() and print it a third time.

	NrFreqs = 1;
	Freq[0] = 10000.0;	Ampl[0] = 1000;

	NrFreqs = MAXFREQS;
// calculate multiplier based on sample frequency and test frequency

	Freq[0] =   100.0;	Ampl[0] = 100;
	Freq[1] =  1000.0;	Ampl[1] = 100;
	Freq[2] = 10000.0;	Ampl[2] = 100;
	Freq[3] = 15000.0;	Ampl[3] = 100;

   DC = 0.0;

// FreqMult = (Freq*360.0/UB.SampleFreq)*(PI/180.0);
	for ( i=0; i<NrFreqs; i++ )			// convert to radians
	 	FreqMult[i] = 2.0*PI*Freq[i]/UB.SampleFreq;

// first calculate floating point values and determine maximum value
   Vmax = 0;
   for (k = 0; k < 2*iSampleSize; k++)
   { 	ADvalues[k] = 0;
	   for ( i=0; i<NrFreqs; i++ )
		{	ftemp = FreqMult[i]*k;			// calculate modulus(2PI)
			j = (long)(ftemp/(2.0*PI));
			ftemp = ftemp - ((float)j * 2.0 * PI);
      	ADvalues[k] += (int)(Ampl[i] * sin(ftemp));
      }
      Vmax = max( Vmax, fabs( ADvalues[k] ) );
   }

#if DEBUG_FFT_DISPLAY & 0x10
   printf("Raw data befor scaling\n");
	for ( i=0; i<2*iSampleSize; i+=4 )
   	printf ( "%4d %6d %6d %6d %6d\n",
      	i, ADvalues[i], ADvalues[i+1], ADvalues[i+2], ADvalues[i+3] );
#endif

    // scale floating point values to be less than 32767 in
    // absolute value and load integer x[k] array.

	fscale = max( 1.0, Vmax/1024.0 );
	for (k = 0; k < 2*iSampleSize; k++)
        ADvalues[k] = (int) ( ADvalues[k] / fscale + 0.5);  // 0.5 rounds
#if DEBUG_FFT_DISPLAY & 0x10
	if (fscale > 1.0 ) printf("input divided by %8.3f\n"  \
                              "to emulate 10 bit A/D\n"\
                              "------------------------\n", fscale);
   printf("Raw data after scaling\n");
	for ( i=0; i<2*iSampleSize; i+=4 )
   	printf ( "%4d %5d %5d %5d %5d\n",
      	i, ADvalues[i], ADvalues[i+1], ADvalues[i+2], ADvalues[i+3] );
#endif
	i=0;											// for breakpoint
  return;
} // DummyData

#endif // USE_DUMMY_DATA	****************************************************


/* START FUNCTION DESCRIPTION ********************************************
TestIO

SYNTAX:			TestIO ( void );

DESCRIPTION:	Exercise the LCD I/O bits
						I/O usage:
							The LCD uses (see F51553.lib):
								PE5 = A0: 1 = data, 0 = command
								PE1 = /ChipSelect
								PE3 = /Reset
		   					PE7 = PWM for BackLight brightness
							The A/D uses:
   							PE0 = /ChipSelect
							The relays use:
								PD0 = K1
								PD2 = K2
   							PD4 = K3
		   					PD6 = K4
							The switches use (open = high)
								PD1 = Sw1
   							PD3 = Sw2
		   					PD5 = Sw3
   							PD7 = Sw4
					This function is meant to be used as a debugging aid and
               should only be exercised when the LCD is NOT plugged into
               the adapter board.  The easiest way to use it is to single
               step through the function and use a voltmeter to monitor
               each of the I/O bits in turn.

					F51553_Init and ADinit must be executed first to initialize
               the ports

PARAMETER:		none

RETURN VALUE:	none

END DESCRIPTION **********************************************************/


debug void TestIO ( void )
{

//*****************************************
// copied from LCDinit
// init the state of the SPI clock on PB0
	BitWrPortI ( PBDR, &PBDRShadow, 1, 0 );
	BitWrPortI ( PBDDR, &PBDDRShadow, 1, 0 );

// init LCD control bits: all = 0
	WrPortI ( LCD_REG, &LCD_SHADOW, 0 );
   BitWrPortI ( LCD_REG, &LCD_SHADOW, 1, LCD_CS_BIT); // make /CS high
	WrPortI ( LCD_DDR, &LCD_DDR_SHADOW, LCD_DDR_SHADOW|LCD_CS_MASK|LCD_A0_MASK|LCD_RESET_MASK|LCD_PWM_MASK );
//*****************************************

// init A/D control bits : A0=1, /CS=1, /Reset=0
	BitWrPortI ( AD_DR, &AD_SHADOW, 1, AD_SHADOW|AD_CS_BIT );// insure A/D is not enabled
	BitWrPortI ( AD_DDR, &AD_DDR_SHADOW, 1, AD_DDR_SHADOW|AD_CS_BIT );// make /CS bit an output
//*****************************************

	while (1)
   {
	   BitWrPortI ( PCDR, &PCDRShadow, 1, 4);	// exercise TxB
   	BitWrPortI ( PCDR, &PCDRShadow, 0, 4);

	   BitWrPortI ( PBDR, &PBDRShadow, 1, 0); // exercise ClkB
   	BitWrPortI ( PBDR, &PBDRShadow, 0, 0);

	   BitWrPortI ( LCD_REG, &LCD_SHADOW, 1, LCD_A0_BIT);
   	BitWrPortI ( LCD_REG, &LCD_SHADOW, 0, LCD_A0_BIT);

	   BitWrPortI ( LCD_REG, &LCD_SHADOW, 1, LCD_RESET_BIT);
   	BitWrPortI ( LCD_REG, &LCD_SHADOW, 0, LCD_RESET_BIT);

	   BitWrPortI ( LCD_REG, &LCD_SHADOW, 1, LCD_PWM_BIT);
   	BitWrPortI ( LCD_REG, &LCD_SHADOW, 0, LCD_PWM_BIT);

   	BitWrPortI ( LCD_REG, &LCD_SHADOW, 0, LCD_CS_BIT);
	   BitWrPortI ( LCD_REG, &LCD_SHADOW, 1, LCD_CS_BIT);

	   BitWrPortI ( AD_DR, &AD_SHADOW, 0, AD_CS_BIT);
   	BitWrPortI ( AD_DR, &AD_SHADOW, 1, AD_CS_BIT);
	}

} // TestIO

